/*
  bscg - an experimental binary-based static call graph generator
         using libbfd, libopcodes and DWARF2-XML.
  Copyright (C) 2003, by gondow@cs.titech.ac.jp, All rights reserved.
  $Id: bscg.c,v 1.33 2004/01/22 12:52:21 gondow Exp gondow $
 */
/* ======================================================================== */
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <time.h>

#include <libxml/parser.h>
#include <libxml/xpath.h>
#include <libxml/tree.h>
#include <libxml/debugXML.h>
#include <libxml/xpathInternals.h>

#include <bfd.h>
#include <dis-asm.h>

#define  UNKNOWN_FILE "????"
#define  UNKNOWN_DIR  "????"
/* ====================================================================== */
static struct {
    char *aout_file;
    char *xml_file;
    FILE *output;
    int n_callees;
    char **callees;
    int n_callers;
    char **callers;
    int n_only_files;
    char **only_files;
    int n_ignore_funcs;
    char **ignore_funcs;
    int more;
    int nolocal;
    int depth;
    char *size;
    char *page;
    int src2html;
    int undef;
    int reloc;
    int sym;
    int bfd;
    int url;
    int verbose;
    char args [1024];
} args = {
    NULL,   /* aout_file */
    NULL,   /* xml_file */
    stdout, /* output */
    0,      /* n_callees */
    NULL,   /* callees */
    0,      /* n_callers */
    NULL,   /* callers */
    0,      /* n_only_files */
    NULL,   /* only_files */
    0,      /* n_ignore_funcs */
    NULL,   /* ignore_funcs */
    0,      /* more */
    0,      /* nolocal */
    1,      /* depth */
    NULL,   /* size */
    NULL,   /* page */
    0,      /* src2html */
    0,      /* undef*/
    0,      /* reloc */
    0,      /* sym */
    0,      /* bfd */
    0,      /* url */
    0,      /* verbose */
    ""      /* args */
};

struct file_info {
    char *filename;
    char *dirname;
    unsigned start_addr;
    unsigned end_addr;
    int  n_funcs;
    struct func_info *funcs; /* all functions defined in the file */
};

struct func_info_list {
    struct func_info *func_info;
    int count;
    struct addr_list *call_addr;
    struct func_info_list *next;  /* NULL terminated linear list */
};

struct func_info {
    char *funcname;
    struct file_info *file_info;    /* file where the function is defined */
    int   lineno;
    unsigned start_addr;
    unsigned end_addr;
    unsigned is_global   :1;        /* globally exported or file scoped */
    unsigned inline_type :2;        /* 0: DW_INL_not_inlined,           *
				     * 1: DW_INL_inlined,               *
				     * 2: DW_INL_declared_not_inlined,  *
				     * 3: DW_INL_declared_inlined       */
    unsigned caller_mark :1;	    /* used by '--callers' option       */
    unsigned callee_mark :1;        /* used by '--callees' option       */
    /* ref and id are used to combine two separate func_info for
       inlined functions. */
    unsigned ref;
    unsigned id;
    struct func_info_list *callers; /* all funcs calling the function   */
    struct func_info_list *callees; /* all funcs called by the function */
};

struct addr_list {
    unsigned addr;
    struct addr_list *next;
};
/* ====================================================================== */
static struct file_info *all_files; /* call graph (array [n_files]) */
static int n_files;

static xmlDocPtr  xml_doc  = NULL;
static xmlNodePtr xml_root = NULL;

static bfd *abfd;
static asection *text;
static long n_symtab, n_reltab;
static asymbol **symtab;  /* symbol table (array [n_symtab]) */
static arelent **reltab;  /* relocation table (array [n_reltab]) */
/* ====================================================================== */
static void *emalloc (size_t size);
static void print_usage (void);
static void process_args (int argc, char *argv []);
static void dump_args (void);
static void print_string_array (char **array, int n);
static void check_next_arg (char **argv);
static void set_args  (char ***array_p, int *num_p, char **argv);
static void init (void);
static void final (void);
static void init_symbol_table (void);
static xmlNodeSetPtr eval_xpath (char *xpath, xmlNodePtr from,
				 void **free_p, int *num);
static void process_file_table (void);
static void process_func_table (xmlNodePtr node, struct file_info *file_info);
static void print_symbol_reloc_table (void);
static void process_disasm_code (void);
static void process_inline_funcs (void);
static void process_outer_funcs (void);
static struct func_info_list *func_is_existing (struct func_info_list *list,
						struct func_info *func_info);
static void register_call (struct func_info *caller_func,
			   unsigned caller_addr,
			   struct func_info *callee_func,
			   unsigned callee_addr);
static void print_file_table (void);
static struct func_info *find_func_by_addr (unsigned addr, int do_exactly);
static struct func_info *find_func_by_id (unsigned id);
static struct func_info *find_func_by_name (const char *funcname,
					    const char *filename);
static const char *find_func_by_symaddr (unsigned target, unsigned vma);
static char * node_name (struct func_info *func, char *buf, size_t size);
static int is_ignored_func (struct func_info *func);
static int is_only_file    (struct file_info *file);
static int is_local_func   (struct func_info *func);
static int is_undef_func   (const char *name);
static int is_reloc_addr   (bfd_vma addr, const char **name);
static void nullify_all_marks (void);
static int is_callers (struct func_info *func);
static int is_callees (struct func_info *func);
static void mark_callers (struct func_info *func, int depth);
static void mark_callees (struct func_info *func, int depth);
static void do_src2html (void);
static char *get_color_name (struct func_info *func);
/* ====================================================================== */
static void *
emalloc (size_t size)
{ 
    void *p = malloc (size);
    if (p == NULL) perror ("malloc");
    memset (p, 0, size); /* for sanitary */
    return p;
}

static void
print_usage (void)
{
    fprintf (stderr, 
	     "usage  : bscg [options] a.out dwarf2.xml\n"
	     "options:\n"
	     "  --nolocal             process global functions only\n"
	     "  --more                emit additional information\n"
	     "  --depth  (num|all)    do '--calle[er]s' recursively up to depth num\n"
	     "                        ('all' means infinity, default is 1)\n"
	     "  --only    f.c,f2.c... process the specified files only\n"
	     "  --ignore  fun,fun2... ignore  the specified functions\n"
	     "  --callees fun,fun2..  process the sepecified func's callees only\n"
	     "  --callers fun,fun2..  process the sepecified func's callers only\n"
	     "  --undef               process used but undefined functions\n"
             "  --reloc               use relocation info. for function search\n"
	     "  --sym                 use symbol info. for function search\n"
	     "  --bfd                 use bfd info. for function search\n"
	     "  --url                 embed URLs to source code\n"
	     "  --src2html            translate sources to htmls\n"
	     "  --size    x,y         drawing size in inches (passed to dot)\n"
	     "  --page    x,y         unit of pagination in inches (passed to dot)\n"
	     "  --verbose             tell what is being doing\n"
	     "  --help                print this usage\n"
	     );
}

static void
process_args (int argc, char *argv [])
{
    int    i;
    char  *p;

    if (argc < 3) {
	print_usage (); exit (1);
    }

    for (i = 1; i < argc; i++) {
	if (argv [i][0] != '-') { /* filenames */
	    if (argc - i < 2) {
		print_usage ();
		exit (1);
	    }
	    args.aout_file = argv [i];
	    args.xml_file  = argv [i + 1];
	    break;
	} else if (!strcmp (argv [i], "--only")) { 
	    set_args (&args.only_files, &args.n_only_files, &argv [i]);
	    i++;
	} else if (!strcmp (argv [i], "--ignore")) { 
	    set_args (&args.ignore_funcs, &args.n_ignore_funcs, &argv [i]);
	    i++;
	} else if (!strcmp (argv [i], "--callers")) { 
	    set_args (&args.callers, &args.n_callers, &argv [i]);
	    i++;
	} else if (!strcmp (argv [i], "--callees")) { 
	    set_args (&args.callees, &args.n_callees, &argv [i]);
	    i++;
	} else if (!strcmp (argv [i], "--more")) { 
	    args.more = 1;
	} else if (!strcmp (argv [i], "--nolocal")) { 
	    args.nolocal = 1;
	} else if (!strcmp (argv [i], "--src2html")) { 
	    args.src2html = 1;
	} else if (!strcmp (argv [i], "--url")) { 
	    args.url = 1;
	} else if (!strcmp (argv [i], "--undef")) { 
	    args.undef = 1;
	} else if (!strcmp (argv [i], "--reloc")) { 
	    args.reloc = 1;
	} else if (!strcmp (argv [i], "--sym")) { 
	    args.sym = 1;
	} else if (!strcmp (argv [i], "--bfd")) { 
	    args.bfd = 1;
	} else if (!strcmp (argv [i], "--depth")) { 
	    check_next_arg (&argv [i]);
	    i++;
	    if (!strcmp (argv [i], "all"))
		args.depth = -1;
	    else
		args.depth = atoi (argv [i]);
	} else if (!strcmp (argv [i], "--size")) { 
	    check_next_arg (&argv [i]);
	    i++;
	    args.size = argv [i];
	} else if (!strcmp (argv [i], "--page")) { 
	    check_next_arg (&argv [i]);
	    i++;
	    args.page = argv [i];
	} else if (!strcmp (argv [i], "--verbose")) { 
	    args.verbose = 1;
	} else if (!strcmp (argv [i], "--help")) {
	    print_usage ();
	    exit (1);
	} else {
	    /* unknown option */
	    fprintf (stderr, "unknown option: %s\n", argv [i]);
	    print_usage ();
	    exit (1);
	}
    }

    p = args.args;
    for (i = 0; i < argc; i++) {
	int len = strlen (argv [i]);
	sprintf (p, "%s ", argv [i]);
	p += len + 1;
    }

    if (args.verbose) {
	dump_args ();
    }
}

static void
dump_args (void)
{
    fprintf (stderr,
	     "args.aout_file = %s\n"
	     "args.xml_file  = %s\n"
	     "args.verbose   = %d\n"
	     "args.more      = %d\n"
	     "args.nolocal   = %d\n"
	     "args.depth     = %d\n"
	     "args.n_only_files = %d\n"
	     "args.n_ignore_funcs = %d\n"
	     "args.n_callees = %d\n"
	     "args.n_callers = %d\n"
	     "args.size      = %s\n"
	     "args.page      = %s\n"
	     "args.args      = %s\n",
	     args.aout_file, args.xml_file,
	     args.verbose, args.more, args.nolocal, args.depth,
	     args.n_only_files, args.n_ignore_funcs,
	     args.n_callees, args.n_callers, 
	     (args.size == NULL ? "(null)" : args.size),
	     (args.page == NULL ? "(null)" : args.page),
	     args.args);

    fprintf (stderr, "only_files:\n");
    print_string_array (args.only_files, args.n_only_files);

    fprintf (stderr, "ignore_funcs:\n");
    print_string_array (args.ignore_funcs, args.n_ignore_funcs);

    fprintf (stderr, "callees:\n");
    print_string_array (args.callees, args.n_callees);

    fprintf (stderr, "callers:\n");
    print_string_array (args.callers, args.n_callers);
}

static void
print_string_array (char **array, int n)
{
    int i;
    for (i = 0; i < n; i++) {
	if (array [i] != NULL) {
	    fprintf (stderr, "[%d] = %s\n", i, array [i]);
	}
    }
}

/* divide (a comma-separated) argv[1] into array */
static void
set_args  (char ***array_p, int *num_p, char **argv)
{
    int j, n = 1;
    char *p = argv [1];

    check_next_arg (argv);

    /* first count functions to be ignored */
    while (( p = strpbrk (p, ",")) != NULL) {
	p++;
	n++;
    }

    *num_p = n;
    *array_p = emalloc (sizeof (char *) * n);

    /* then, set them to args.ignore_funcs */
    p = argv [1];
    for (j = 0; j < n; j++) {
	(*array_p) [j] = p;
	p = strpbrk (p, ":");
	if (p ==  NULL) break;
	*p = '\0';
	p++;
    }
}

static void
check_next_arg (char **argv)
{
    if (argv [1] == NULL || argv [1][0] == '-') {
	fprintf (stderr, "no value for the option %s\n", argv [0]);
	print_usage ();
	exit (1);
    }
}

static void
init (void)
{
    char *xml = args.xml_file;

    xmlDoValidityCheckingDefaultValue = 1;
    if (!strcmp (".gz", xml + strlen (xml) -3)) { /* gzipped file */
	FILE *fp;
	char buf[BUFSIZ], *gunzip_file;

	gunzip_file = tempnam (".", NULL);
	snprintf (buf, BUFSIZ, "gunzip -c %s > %s", xml, gunzip_file);
	if (system (buf) == -1) {
	    perror ("system"); exit (1);
	}
	fp = fopen (gunzip_file, "r");
	xml_doc  = xmlParseFile(gunzip_file);
	unlink (gunzip_file);
    } else { /* not gzipped file */
	xml_doc  = xmlParseFile(xml);
    }
    xml_root = xmlDocGetRootElement (xml_doc);
    xmlXPathInit();

    init_symbol_table ();
}

static void
final (void)
{
    bfd_close (abfd);
}

static xmlNodeSetPtr 
eval_xpath (char *xpath, xmlNodePtr from, void **free_p, int *num)
{
    xmlXPathObjectPtr  result;
    xmlXPathContextPtr context;
    context = xmlXPathNewContext(xml_doc);
    assert (context != NULL);
    context->node = from;
    result = xmlXPathEval (xpath, context);
    xmlXPathFreeContext (context);
    if (args.verbose)
	xmlXPathDebugDumpObject (stdout, result, 0);
    if (num != NULL)
	*num = xmlXPathNodeSetGetLength (result->nodesetval);
    *free_p = result;
    return result->nodesetval;
}

static void
init_symbol_table (void)
{
    long size;

    bfd_init ();
    abfd = bfd_openr (args.aout_file, "default");
    if (abfd == NULL) {
	fprintf (stderr, "Cannot open file: %s\n", args.aout_file);
	exit (1);
    }
    bfd_check_format (abfd, bfd_object);
    text = bfd_get_section_by_name (abfd, ".text");
    size = bfd_get_symtab_upper_bound (abfd);
    symtab = emalloc (size);
    n_symtab = bfd_canonicalize_symtab (abfd, symtab);

    size = bfd_get_reloc_upper_bound (abfd, text);
    reltab = emalloc (size);
    n_reltab = bfd_canonicalize_reloc (abfd, text, reltab, symtab);
}

/* setup call_graph.all_files */
static void
process_file_table (void)
{
    int  i;
    void *free_p;
    char *xpath =
	"//section[@name='.debug_info']//tag[@name='DW_TAG_compile_unit']";
    xmlNodeSetPtr nodeset = eval_xpath (xpath, xml_root, &free_p, &n_files);
    /* +1 for a virtual file for undefined functions */
    all_files = emalloc (sizeof (struct file_info) * (n_files + 1));

    for (i = 0; i < n_files; i++) {
	xmlNodePtr p;

	assert (nodeset->nodeTab [i]->children != NULL);

	for (p = nodeset->nodeTab [i]->children; p != NULL; p = p->next) {
	    char *attrname, *value;

	    if (p->type != XML_ELEMENT_NODE) continue;

	    all_files [i].funcs = NULL;

	    attrname = xmlGetProp (p, "name");
	    value    = xmlGetProp (p, "value");

	    if (!strcmp (attrname, "DW_AT_name")) {
		all_files [i].filename = value;
	    } else if (!strcmp (attrname, "DW_AT_comp_dir")) {
		all_files [i].dirname  = value;
	    } else {
		if (!strcmp (attrname, "DW_AT_low_pc")) {
		    all_files [i].start_addr = (unsigned) atoi (value);
		} else if (!strcmp (attrname, "DW_AT_high_pc")) {
		    all_files [i].end_addr   = (unsigned) atoi (value);
		}
		free (value);
	    }
	    free (attrname);
    	}
	process_func_table (nodeset->nodeTab [i], &all_files [i]);
    }

    if (args.sym || args.reloc) {
	all_files [n_files].filename   = UNKNOWN_FILE;
	all_files [n_files].dirname    = UNKNOWN_DIR;
	all_files [n_files].start_addr = 0;
	all_files [n_files].end_addr   = 0;
	all_files [n_files].n_funcs    = 0;
	all_files [n_files].funcs      = NULL;
	n_files++;
    } else {
	all_files [n_files].filename   = NULL; 
    }

    free (free_p); 
}

static void
process_func_table (xmlNodePtr node, struct file_info *file_info)
{
    int  i, j;
    void *free_p1, *free_p2;
    char *xpath1 = "tag[@name='DW_TAG_subprogram']";
    char *xpath2 = "tag[@name='DW_TAG_subprogram'][attribute/@name='DW_AT_declaration']";
    xmlNodeSetPtr nodeset1 = eval_xpath (xpath1, node, &free_p1, NULL);
    xmlNodeSetPtr nodeset2 = eval_xpath (xpath2, node, &free_p2, NULL);
    xmlNodeSetPtr nodeset  = xmlXPathDifference (nodeset1, nodeset2);

    file_info->n_funcs = xmlXPathNodeSetGetLength (nodeset);
    file_info->funcs   = emalloc (sizeof (struct func_info) * file_info->n_funcs);

    for (i = 0; i < file_info->n_funcs; i++) {
	xmlNodePtr p;
	struct func_info *func = &file_info->funcs [i];

	assert (nodeset->nodeTab [i]->children != NULL);

	func->funcname     = NULL;
	func->file_info    = NULL;
	func->lineno       = 0;
	func->start_addr   = 0;
	func->end_addr     = 0;
	func->is_global    = 0;
	func->inline_type  = 0;
	func->caller_mark  = 0;
	func->callee_mark  = 0;
	func->ref          = 0;
	func->id           = 0;
	func->callers      = NULL;
	func->callees      = NULL;

	for (p = nodeset->nodeTab [i]->children; p != NULL; p = p->next) {
	    char *attrname, *value;

	    if (p->type != XML_ELEMENT_NODE) continue;

	    attrname = xmlGetProp (p, "name");
	    value    = xmlGetProp (p, "value");

	    if (!strcmp (attrname, "DW_AT_name")) {
		func->funcname = value;
	    } else if (!strcmp (attrname, "DW_AT_decl_file")) {
		func->file_info = file_info;
	    } else {
		if (!strcmp (attrname, "DW_AT_decl_line")) {
		    func->lineno = atoi (value);
		} else if (!strcmp (attrname, "DW_AT_low_pc")) {
		    func->start_addr = atoi (value);
		} else if (!strcmp (attrname, "DW_AT_high_pc")) {
		    func->end_addr = atoi (value);
		} else if (!strcmp (attrname, "DW_AT_external")) {
		    func->is_global = atoi (value);
		    assert (func->is_global == 1);
		} else if (!strcmp (attrname, "DW_AT_declaration")) {
		    /* funcs with DW_AT_declaration are all excluded
		       before creating function tables, since they are
		       `exterinal' declarations. */
		    assert (0);
		} else if (!strcmp (attrname, "DW_AT_inline")) {
		    char *offset = xmlGetProp (nodeset->nodeTab [i], "offset");
		    func->inline_type = atoi (value);
		    func->id = atoi (offset + 5); /* +5 to skip "info:" */
		    free (offset);
		} else if (!strcmp (attrname, "DW_AT_abstract_origin")) {
		    func->ref = atoi (value);
		}
		free (value);
	    }
	    free (attrname);
	}
    }
    free (free_p1);
    free (free_p2);

    /* To cope with inlined functions, any ref/id pair should be
       merged into one func_info entity here by eliminating ref. */
    for (i = 0; i < file_info->n_funcs; i++) {
	struct func_info *func = &file_info->funcs [i];

	if (func->ref == 0) continue;

	for (j = 0; j < file_info->n_funcs; j++) {
	    struct func_info *func2 = &file_info->funcs [j];
	    if (func->ref == func2->id) {
		func2->start_addr = func->start_addr;
		func2->end_addr = func->end_addr;
		/* this NULL implies that we have to skip NULL entries
		   for file_info interations */
		func->funcname = NULL;
	    }
	}
    }
}

static void
print_symbol_reloc_table (void)
{
    int i;

    for (i = 0; i < n_symtab; i++) {
	symbol_info syminfo;
	bfd_symbol_info (symtab [i], &syminfo);
	if (syminfo.name != NULL) {
	    printf ("%c, %x, %s\n", 
		    syminfo.type, (unsigned) syminfo.value, syminfo.name);
	}
    }

    for (i = 0; i < n_reltab; i++) {
	if (reltab [i]->sym_ptr_ptr == NULL) continue;
	if (*reltab [i]->sym_ptr_ptr == NULL) continue;
	if ((*reltab [i]->sym_ptr_ptr)->name == '\0') continue;
	printf ("sym_ptr_ptr = %s, address = %x, addend = %x\n",
		(*reltab [i]->sym_ptr_ptr)->name,
		(unsigned)reltab [i]->address,
		(unsigned)reltab [i]->addend);
    }
}

static void
process_disasm_code (void)
{
    FILE *fp_null;
    struct disassemble_info disasm_info;
    disassembler_ftype disassemble_fn;
    int insn_len;

    bfd_byte *data;
    bfd_size_type datasize;
    bfd_vma  p;

    /* for disassembler to dispose all output to /dev/null */
    if ((fp_null = fopen ("/dev/null", "w")) == NULL) {
	perror ("fopen"); exit (1);
    }

    INIT_DISASSEMBLE_INFO (disasm_info, fp_null, fprintf);
    disasm_info.flavour = bfd_get_flavour (abfd);
    disasm_info.arch = bfd_get_arch (abfd);
    disasm_info.mach = bfd_get_mach (abfd);
    disasm_info.disassembler_options = NULL;
    disasm_info.octets_per_byte = bfd_octets_per_byte (abfd);

    if (bfd_big_endian (abfd)) {
	disasm_info.display_endian = disasm_info.endian = BFD_ENDIAN_BIG;
    } else if (bfd_little_endian (abfd)) {
	disasm_info.display_endian = disasm_info.endian = BFD_ENDIAN_LITTLE;
    } else {
	assert (0);
    }

    datasize = bfd_get_section_size_before_reloc (text);
    data = emalloc (datasize);
    bfd_get_section_contents (abfd, text, data, 0, datasize);

    disasm_info.buffer = data;
    disasm_info.buffer_vma = text->vma;
    disasm_info.buffer_length = datasize;
    disasm_info.section = text;

    disassemble_fn = disassembler (abfd);

    disasm_info.insn_info_valid = 0; /* supported by only few architectures */

    for (p = text->vma; p < text->vma + datasize; p += insn_len) {
	insn_len = disassemble_fn (p, &disasm_info);
	assert (disasm_info.insn_info_valid);

	switch (disasm_info.insn_type) {
	case dis_jsr:	   /* jump to subroutine */
	case dis_condjsr:  /* conditional jump to subroutine */
	{
	    const char *caller_name, *callee_name, *caller_file, *callee_file;
	    unsigned lineno;
	    struct func_info *callee, *caller;

	    if (args.reloc && is_reloc_addr (p, &callee_name)) {
		callee = find_func_by_name (callee_name, NULL);
		if (args.bfd) {
		    bfd_find_nearest_line (abfd, text, symtab, p - text->vma,
					   &caller_file, &caller_name, &lineno);
		    caller = find_func_by_name (caller_name, caller_file);
		    register_call (caller, NULL, callee, NULL);
		} else {
		    register_call (NULL, (unsigned) p, callee, NULL);
		}
	    } else if (args.bfd) {
		/* bfd_find_nearest_line returns some wrong name for
		   dynamically-linked functions like `printf'. */
#if 1
		callee_name = find_func_by_symaddr (disasm_info.target,
						    text->vma);
#else
		bfd_find_nearest_line (abfd, text, symtab,
				       disasm_info.target - text->vma,
				       &callee_file, &callee_name, &lineno);
#endif
		if (callee_name == NULL) continue;
		callee = find_func_by_name (callee_name, NULL);

		bfd_find_nearest_line (abfd, text, symtab, p - text->vma,
				       &caller_file, &caller_name, &lineno);
		caller = find_func_by_name (caller_name, NULL);
#if 0 /* for debug */
		printf ("call: %s -> %s\n", caller_name, callee_name);
#endif
		register_call (caller, NULL, callee, NULL);
	    } else {
		register_call (NULL, (unsigned) p,
			       NULL, (unsigned)disasm_info.target);
	    }
	}
	default:
	    /* the other insns are just ignored */
	    break;
	}
    }
}

static void
process_inline_funcs (void)
{
    int  i, n;
    void *free_p;
    char *xpath = "//section[@name='.debug_info']"
	    	  "//tag[@name='DW_TAG_inlined_subroutine']";
    xmlNodeSetPtr nodeset = eval_xpath (xpath, xml_root, &free_p, &n);

    for (i = 0; i < n; i++) {
	xmlNodePtr p;
	unsigned low_pc = 0, id = 0;
	struct func_info *func;

	assert (nodeset->nodeTab [i]->children != NULL);
	for (p = nodeset->nodeTab [i]->children; p != NULL; p = p->next) {
	    char *attrname, *value;

	    if (p->type != XML_ELEMENT_NODE) continue;
	    attrname = xmlGetProp (p, "name");
	    value    = xmlGetProp (p, "value");

	    if (!strcmp (attrname, "DW_AT_low_pc")) {
		low_pc = atoi (value);
	    } else if (!strcmp (attrname, "DW_AT_abstract_origin")) {
		id     = atoi (value);
	    }
	    free (attrname); free (value);
    	}

	assert (low_pc != 0);
	func = find_func_by_id (id);
	assert (func != NULL);
	register_call (NULL, low_pc, func, 0);
    } 
    free (free_p); 
}

static void
process_outer_funcs (void)
{
    int i, ret, n_funcs;
    struct outer_func {
	unsigned addr;
	const char *funcname;
	struct outer_func *next;
    } *outer_func_list = NULL, **pp, **qq;

    if (!args.reloc && !args.sym)
	return;

    /* collect outer functions from reloc. table */
    if (!args.reloc)
	goto end_reloc;
    for (i = 0; i < n_reltab; i++) {
	struct symbol_cache_entry **sym_entry;

	if (strcmp (reltab [i]->howto->name, "R_SPARC_WDISP30")) continue;
	sym_entry = reltab [i]->sym_ptr_ptr;
	if (sym_entry == NULL || *sym_entry == NULL) continue;
	if ((*sym_entry)->name == '\0') continue;

	if (!is_undef_func ((*sym_entry)->name)) continue;

	/* insert this (undefiend) function reloc info into outer_func_list */
	for (pp = &outer_func_list; ; pp = &(*pp)->next) {
	    if ((*pp == NULL)
		|| (ret = strcmp ((*sym_entry)->name, (*pp)->funcname)) < 0) {
		/* insert this function in the increasing order */
		struct outer_func *func;
		func = emalloc (sizeof (struct outer_func));
		func->addr = (unsigned)reltab [i]->address,
		func->funcname = (*sym_entry)->name;
		func->next = *pp;
		*pp = func;
		break;
	    } else if (ret == 0) {
		break;	  /* make sure that there is no duplication */
	    }
	}
    }
end_reloc:

    /* collect outer functions from sym. table */
    if (!args.sym)
	goto end_sym;
    for (i = 0; i < n_symtab; i++) {
	symbol_info syminfo;
	bfd_symbol_info (symtab [i], &syminfo);

	if (syminfo.type != 'T' && syminfo.type != 'U')     continue;
	if (find_func_by_name (syminfo.name, NULL) != NULL) continue;

	/* insert this (undefiend) function sym info into outer_func_list */
	for (pp = &outer_func_list; ; pp = &(*pp)->next) {
	    if ((*pp == NULL)
		|| (ret = strcmp (syminfo.name, (*pp)->funcname)) < 0) {
		/* insert this function in the increasing order */
		struct outer_func *func;
		func = emalloc (sizeof (struct outer_func));
		func->addr =     syminfo.value;
		func->funcname = syminfo.name;
		func->next = *pp;
		*pp = func;
		break;
	    } else if (ret == 0) {
		break;	  /* make sure that there is no duplication */
	    }
	}
    }
end_sym:

    /* count the number of elements in outer_func_list */
    n_funcs = 0;
    for (pp = &outer_func_list; *pp != NULL; pp = &(*pp)->next)
	n_funcs++;

    /* make entries in all_files [n_files - 1] */
    all_files [n_files - 1].n_funcs = n_funcs;
    all_files [n_files - 1].funcs   = emalloc (sizeof (struct func_info)
					       * n_funcs);

    i = 0;
    for (pp = &outer_func_list; *pp != NULL; pp = &(*pp)->next) {
	struct func_info *func = &all_files [n_files - 1].funcs [i];

	func->funcname     = (char *)(*pp)->funcname;
	func->file_info    = &all_files [n_files - 1];
	func->lineno       = 0;
	func->start_addr   = 0;
	func->end_addr     = 0;
	func->is_global    = 1;
	func->inline_type  = 0;
	func->caller_mark  = 0;
	func->callee_mark  = 0;
	func->ref          = 0;
	func->id           = 0;
	func->callers      = NULL;
	func->callees      = NULL;

	i++;
    }

    /* free outer_func_list */
    for (pp = &outer_func_list; *pp != NULL; pp = qq) {
	qq = &(*pp)->next;
	free (*pp);
    }
}

static void
register_call (struct func_info *caller_func, unsigned caller_addr,
	       struct func_info *callee_func, unsigned callee_addr)
{
    struct func_info_list *list1, *list2, *ret;
    struct addr_list *addr_list1, *addr_list2, **p;

    if (caller_func == NULL)
	caller_func = find_func_by_addr (caller_addr, 0);
    if (callee_func == NULL)
	callee_func = find_func_by_addr (callee_addr, 1);

    if (caller_func == NULL || callee_func == NULL)
	return;

#if 0	 /* for debugging */
    fprintf (stderr, "===================\n");
    fprintf (stderr, "caller = %s\n", caller_func->funcname);
    fprintf (stderr, "caller addr = %x\n", caller_addr);
    fprintf (stderr, "callee = %s\n", callee_func->funcname);
    fprintf (stderr, "callee addr = %x\n", callee_addr);
#endif

    ret = func_is_existing (caller_func->callees, callee_func);
    if (ret == NULL) {
	list1 = emalloc (sizeof (struct func_info_list));
	list1->func_info = callee_func;
	list1->count     = 1;
	list1->call_addr = NULL;
	list1->next      = caller_func->callees;
	caller_func->callees   = list1;
    } else {
	ret->count++;
	list1 = ret;
    }

    /* insert new struct addr_list in increasing order */
    for (p = &list1->call_addr; ; p = &(*p)->next) {
	if (*p == NULL || (*p)->addr > caller_addr) {
	    addr_list1 = emalloc (sizeof (struct addr_list));
	    addr_list1->addr = caller_addr;
	    addr_list1->next = *p;
	    *p = addr_list1;
	    break;
	}
    }

    ret = func_is_existing (callee_func->callers, caller_func);
    if (ret == NULL) {
	list2 = emalloc (sizeof (struct func_info_list));
	list2->func_info = caller_func;
	list2->count     = 1;
	list2->call_addr = NULL;
	list2->next      = callee_func->callers;
	callee_func->callers   = list2;
    } else {
	ret->count++;
	list2 = ret;
    }

    for (p = &list2->call_addr; ; p = &(*p)->next) {
	if (*p == NULL || (*p)->addr > caller_addr) {
	    addr_list2 = emalloc (sizeof (struct addr_list));
	    addr_list2->addr = caller_addr;
	    addr_list2->next = *p;
	    *p = addr_list2;
	    break;
	}
    }
}

static struct func_info_list *
func_is_existing (struct func_info_list *list, struct func_info *func_info)
{
    for (; list != NULL; list = list->next) {
	if (list->func_info == func_info)
	    return list;
    }
    return NULL;
}

static void
print_file_table (void)
{
    int i, j;

    for (i = 0; i < n_files; i++) {
	struct file_info *file = &all_files [i];
	fprintf (stderr, "file = %s, dir = %s\n",
		 file->filename, file->dirname);

	for (j = 0; j < file->n_funcs; j++) {
	    struct func_info *func = &file->funcs [j];
	    if (func->funcname == NULL) continue; /* skip NULL entry */
	    fprintf (stderr, "funcname = %s\n", func->funcname);
	}
    }
}

static struct func_info *
find_func_by_addr (unsigned addr, int do_exactly)
{
    int i, j;

    for (i = 0; i < n_files; i++) {
	struct file_info *file = &all_files [i];
	if (addr < file->start_addr || addr >= file->end_addr) 
	    continue;
	for (j = 0; j < file->n_funcs; j++) {
	    struct func_info *func = &file->funcs [j];
	    if (func->funcname == NULL) continue; /* skip NULL entry */
	    if (addr < func->start_addr || addr >= func->end_addr)
		continue;
	    if (do_exactly) {
		if (addr == func->start_addr) 
		    return func;
		else
		    return NULL;
	    }
	    return func;
	}
    }
    return NULL;
}

static struct func_info *
find_func_by_id (unsigned id)
{
    int i, j;

    for (i = 0; i < n_files; i++) {
	struct file_info *file = &all_files [i];

	for (j = 0; j < file->n_funcs; j++) {
	    struct func_info *func = &file->funcs [j];
	    if (func->funcname == NULL) continue; /* skip NULL entry */

	    if (func->id == id)
		return func;
	}
    }
    return NULL;
}

static struct func_info *
find_func_by_name (const char *funcname, const char *filename)
{
    int i, j;

    for (i = 0; i < n_files; i++) {
	struct file_info *file = &all_files [i];
	if (filename != NULL && strcmp (filename, file->filename))
	    continue;
	for (j = 0; j < file->n_funcs; j++) {
	    struct func_info *func = &file->funcs [j];
	    if (func->funcname == NULL) continue; /* skip NULL entry */
	    if (!strcmp (funcname, func->funcname))
		return func;
	}
    }
    return NULL;
}

static const char *
find_func_by_symaddr (unsigned target, unsigned vma)
{
    int i;
    unsigned addr;

    for (i = 0; i < n_symtab; i++) {
	symbol_info syminfo;
	bfd_symbol_info (symtab [i], &syminfo);
	if (syminfo.type == 'U')
	    addr = target;
	else
	    addr = target - vma;
	    
	if (addr == symtab [i]->value) {
	    return symtab [i]->name;
	}
    }
    return NULL;
}

static void
output_call_graph_by_dot (void)
{
    int i, j;
    time_t now;
    char *p, *now_str;
    FILE *fp = args.output;
    char buf [BUFSIZ], buf2 [BUFSIZ];
    struct file_info *file;
    struct func_info *func;
    struct func_info_list *list;

    getcwd (buf, sizeof (buf));
    now = time (NULL);
    now_str = ctime (&now);
    p = strrchr (now_str, '\n');
    *p = '\0'; /* eleminate '\n' */
    fprintf (fp, "digraph callgraph {\n"
	     "\tlabel=\"%s\\n%% %s\\n%s\\n%s\";\n"
	     "\tnode  [fontname=Helvetica,style=filled,shape=box];\n"
	     "\tedge  [fontname=Helvetica];\n"
	     "\tgraph [fontname=Helvetica];\n"
	     "\trankdir=LR;\n", args.aout_file, args.args, buf, now_str);
    if (args.size != NULL)
	fprintf (fp, "\tsize=\"%s\";\n", args.size);
    if (args.page != NULL)
	fprintf (fp, "\tpage=\"%s\";\n", args.page);

    if (args.callers != NULL || args.callees != NULL) {
	/* prepare to process '--callers' and/or '--callees' options. */
	nullify_all_marks ();

	/* now marking */
	for (i = 0; i < n_files; i++) {
	    file = &all_files [i];
	    for (j = 0; j < file->n_funcs; j++) {
		func = &file->funcs [j];
		if (func->funcname == NULL) continue;
		if (is_callers (func)) mark_callers (func, args.depth);
		if (is_callees (func)) mark_callees (func, args.depth);
	    }
	}
    }

    /* output */
    for (i = 0; i < n_files; i++) {
	file = &all_files [i];
	if (!is_only_file (file)) continue;
	if (file->filename == NULL) continue;
	if ((args.sym || args.reloc) && !args.undef && (i == n_files - 1))
	    continue;

	fprintf (fp, "\n");

	if (args.more) {
	    fprintf (fp, "\tsubgraph cluster%d {\n"
		     "\t\tlabel=\"%s\\n%s\\n(%x-%x)\";\n",
		     i, file->filename, file->dirname,
		     file->start_addr, file->end_addr);
	} else {
	    fprintf (fp, "\tsubgraph cluster%d {\n"
		     "\t\tlabel=\"%s\";\n", i, file->filename);
	}

	/* output node */
	for (j = 0; j < file->n_funcs; j++) {
	    func = &file->funcs [j];
	    if (func->funcname == NULL) continue;

	    /* filtering */
	    if (is_ignored_func (func))           continue;
	    if (args.nolocal && is_local_func (func)) continue;
	    if (args.callers != NULL && func->caller_mark == 0) continue;
	    if (args.callees != NULL && func->callee_mark == 0) continue;

	    fprintf (fp, "\t\t%s [color=%s,peripheries=%d,",
		     node_name (func, buf, sizeof (buf)),
		     get_color_name (func),
		     func->is_global + 1);
	    if (args.more) {
		fprintf (fp, "label=\"%s\\n%s:%d\\n(%x-%x)\",",
			 func->funcname, 
			 func->file_info->filename,
			 func->lineno,
			 func->start_addr, func->end_addr);
	    } else {
		fprintf (fp, "label=%s,", func->funcname);
	    }
	    if (args.url) {
		fprintf (fp, "URL=\"%s.html#%d\",", 
			 file->filename, func->lineno - 5);
	    }
	    fprintf (fp, "];\n");
	}
	fprintf (fp, "\t}\n\n");
	    
	/* output edge */
	for (j = 0; j < file->n_funcs; j++) {
	    func = &file->funcs [j];
	    if (func->funcname == NULL) continue;

	    /* filtering */
	    if (is_ignored_func (func))                         continue;
	    if (args.nolocal && is_local_func (func))           continue;
	    if (args.callers != NULL && func->caller_mark == 0) continue;
	    if (args.callees != NULL && func->callee_mark == 0) continue;

	    for (list = func->callees; list != NULL; list = list->next) {
		char *caller = node_name (func, buf, sizeof (buf));
		char *callee = node_name (list->func_info,
					  buf2, sizeof (buf2));
		struct func_info *func2 = list->func_info;

		/* filtering */
		if (is_ignored_func (func2))                         continue;
		if (args.nolocal && !func2->is_global)               continue;
		if (args.callers != NULL && func2->caller_mark == 0) continue;
		if (args.callees != NULL && func2->callee_mark == 0) continue;

		fprintf (fp, "\t\t%s->%s[", caller, callee);
		if (args.more && list->count > 1)
		    fprintf (fp, "label=%d,", list->count);
		if (args.url) {
		    const char *filename, *funcname;
		    unsigned lineno;
		    bfd_find_nearest_line (abfd, text, symtab,
					   list->call_addr->addr - text->vma,
					   &filename, &funcname, &lineno);
		    fprintf (fp, "URL=\"%s.html#%d\",", 
			     file->filename, lineno);
		}
		if (func2->inline_type == 1 || func2->inline_type == 3) {
		    fprintf (fp, "style=dashed,");
		}
		fprintf (fp, "];\n");

		if (!is_only_file (list->func_info->file_info)) {
		    fprintf (fp, "\t\t%s[label=\"%s\\n%s:%d\"];\n", 
			     callee,
			     list->func_info->funcname,
			     list->func_info->file_info->filename,
			     list->func_info->lineno);
		}
	    }
	}
    }

    fprintf (fp, "}\n");
}

static char *
get_color_name (struct func_info *func)
{
    /* inline_type 1: not inline-declared, but inlined,
       inline_type 3: inline-declared, and inlined */
    if (func->inline_type == 1 || func->inline_type == 3)
	return "greenyellow";
    else if (func->is_global)
	return "lightpink";
    else
	return "lightblue";
}

static char *
node_name (struct func_info *func, char *buf, size_t size)
{
    if (func->is_global) {
	snprintf (buf, size, "\"%s\"", func->funcname);
    } else {
	snprintf (buf, size, "\"%s:%s\"", func->file_info->filename,
		  func->funcname);
    }
    return buf;
}

static int
is_ignored_func (struct func_info *func)
{
    int i;
    for (i = 0; i < args.n_ignore_funcs; i++) {
	if (!strcmp (func->funcname, args.ignore_funcs [i]))
	    return 1;
    }
    return 0;
}

static int
is_only_file (struct file_info *file)
{
    int i;

    if (args.n_only_files == 0)
	return 1;

    for (i = 0; i < args.n_only_files; i++) {
	if (!strcmp (file->filename, args.only_files [i]))
	    return 1;
    }
    return 0;
}

/* This function is used when args.nolocal is set. The point is that
   a local function should be output when one of its callees in the
   other files is global. */
static int
is_local_func (struct func_info *func)
{
    struct func_info_list *p;

    if (func->is_global)
	return 0;

    for (p = func->callees; p != NULL; p = p->next) {
	if (p->func_info->is_global
	    && strcmp (p->func_info->file_info->filename,
		       func->file_info->filename))
	    return 0;
    }
    return 1;
}

static int
is_reloc_addr   (bfd_vma addr, const char **name)
{
    int i;
    for (i = 0; i < n_reltab; i++) {
	struct symbol_cache_entry **sym_entry = reltab [i]->sym_ptr_ptr;

	if (strcmp (reltab [i]->howto->name, "R_SPARC_WDISP30")) continue;

	if (reltab [i]->address == addr) {
	    *name = (char *)(*sym_entry)->name;
	    return 1;
	}
    }
    *name = NULL;
    return 0;
}

static int
is_undef_func   (const char *name)
{
    int i;
    symbol_info syminfo;
    assert (name != NULL);

    for (i = 0; i < n_symtab; i++) {
	bfd_symbol_info (symtab [i], &syminfo);
	if (syminfo.name != NULL && !strcmp (name, syminfo.name))
	    return syminfo.type == 'U';
    }
    fprintf (stderr, "no such symbol: %s\n", syminfo.name);
    return 1;
}

static void
nullify_all_marks (void)
{
    int i, j;

    for (i = 0; i < n_files; i++) {
	struct file_info *file = &all_files [i];

	for (j = 0; j < file->n_funcs; j++) {
	    struct func_info *func = &file->funcs [j];
	    func->caller_mark = NULL;
	    func->callee_mark = NULL;
	}
    }
}

static int
is_callers (struct func_info *func)
{
    int i;
    for (i = 0; i < args.n_callers; i++) {
	if (!strcmp (args.callers [i], func->funcname))
	    return 1;
    }
    return 0;
}

static int
is_callees (struct func_info *func)
{
    int i;
    for (i = 0; i < args.n_callees; i++) {
	if (!strcmp (args.callees [i], func->funcname))
	    return 1;
    }
    return 0;
}

/* depth == -1 means infinite large number */
static void
mark_callers (struct func_info *func, int depth)
{
    struct func_info_list *list;

    if (depth == -1 && func->caller_mark)
	return; /* to avoid infinite recursion */ 

    func->caller_mark = 1; /* mark it! */

    if (depth == 0) return;
    if (depth >  0) depth--;
    for (list = func->callers; list != NULL; list = list->next)
	mark_callers (list->func_info, depth);
}

static void
mark_callees (struct func_info *func, int depth)
{
    struct func_info_list *list;

    if (depth == -1 && func->callee_mark)
	return; /* to avoid infinite recursion */

    func->callee_mark = 1; /* mark it! */

    if (depth == 0) return;
    if (depth >  0) depth--;
    for (list = func->callees; list != NULL; list = list->next)
	mark_callees (list->func_info, depth);
}

static void
do_src2html (void)
{
    int i, ret;
    struct file_info *file;
    char command [1024];

    for (i = 0; i < n_files; i++) {
	file = &all_files [i];
	snprintf (command, sizeof (command), "PATH=.:$PATH; src2html %s",
		  file->filename);
#if 0
	snprintf (command, sizeof (command), "PATH=.:$PATH; src2html %s/%s",
		  file->dirname, file->filename);
#endif
	ret = system (command);
	if (ret == -1) {
	    perror ("system"); exit (1);
	}
    }
}
/* ====================================================================== */
int
main (int argc, char *argv [])
{
    process_args (argc, argv);
    init ();

    /* building call graph information */
    process_file_table ();
#if 0
    print_file_table ();
    //print_symbol_reloc_table ();
#endif
    process_outer_funcs ();
    process_disasm_code ();
    process_inline_funcs ();

    if (args.src2html)
	do_src2html ();

    /* output call graph information */
    output_call_graph_by_dot ();

    final ();
    return 0;
}
/* ====================================================================== */

/*
  ꤷץ:

  饤֥ꤹ? Ƥ椫饤֥Ǥ?

    DOMĥ꡼Υ󤷤(--only Ƥ⾮ʤʤ)
    .rela.debug_info (debug_* κ־μ谷)
    ֥ȥե뤿ᡤ־Ȥ(--reloc)
    extern ΤߤǼΤΤʤؿɽ롥(--undef)

done  饤ؿб
done  URL ץ (ɤ?)
done ꤷؿ callees or calleers  transitive closures ɽ롥
done  ꤷؿ callees or callers ɽ롥
done  ꤷեδؿɽ롥
done  Хؿɽ롥
done  ꤷؿɽ鳰
done  verbose -- ֹ桤ǥ쥯ȥꡤɥ쥹ʤɤξ񤯡
done  ե륹״ؿϡ֥ե̾:ؿ̾פˤ롥
done  callee  exact address Τߤ褦ˤ롥ʤȡ
done  GOT ΤΡcall ̿⥲åȤƤޤ
 */
